{% extends "epic/base_layout.html" %}
{% load crispy_forms_tags %}
{% load staticfiles %}

{% block extrahead %}
{{ block.super }}
{{ form.media }}
{% endblock extrahead %}

{% block content %}
{{ block.super }}


<div class="panel panel-default">
  <div class="panel-heading">
    Compare bill-of-materials (BOM) of part {{part.html_link|safe}}
    against BOM of:
  </div>
  <div class="panel-body">
    <form method="POST">
      {% crispy form %}
      <div class="form-actions">
	<input type="submit" name="cmp-qty" value="Compare Quantity"
		class="btn btn-primary">
	<input type="submit" name="cmp-refdes" value="Compare Refdes"
		class="btn btn-secondary">
      </div>
    </form>
  </div>
</div>

{% endblock content %}
